import grpc
from github.com.metaprov.modelaapi.pkg.apis.infra.v1alpha1.generated_pb2 import VirtualVolume as MDVirtualVolume
from github.com.metaprov.modelaapi.services.virtualvolume.v1.virtualvolume_pb2_grpc import VirtualVolumeServiceStub
from github.com.metaprov.modelaapi.services.virtualvolume.v1.virtualvolume_pb2 import CreateVirtualVolumeRequest, \
    UpdateVirtualVolumeRequest, \
    DeleteVirtualVolumeRequest, GetVirtualVolumeRequest, ListVirtualVolumesRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class VirtualVolume(Resource):
    def __init__(self, item: MDVirtualVolume = MDVirtualVolume(), client=None, namespace="", name=""):
        super().__init__(item, client, namespace=namespace, name=name)


class VirtualVolumeClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: VirtualVolumeServiceStub = stub

    def create(self, virtualvolume: VirtualVolume) -> bool:
        request = CreateVirtualVolumeRequest()
        request.virtualvolume.CopyFrom(virtualvolume.raw_message)
        try:
            response = self.__stub.CreateVirtualVolume(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, virtualvolume: VirtualVolume) -> bool:
        request = UpdateVirtualVolumeRequest()
        request.virtualvolume.CopyFrom(virtualvolume.raw_message)
        try:
            self.__stub.UpdateVirtualVolume(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[VirtualVolume, bool]:
        request = GetVirtualVolumeRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetVirtualVolume(request)
            return VirtualVolume(response.virtualvolume, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteVirtualVolumeRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteVirtualVolume(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[VirtualVolume], bool]:
        request = ListVirtualVolumesRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListVirtualVolumes(request)
            return [VirtualVolume(item, self) for item in response.virtualvolumes.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import time

from google.protobuf.message import Message
from k8s.io.apimachinery.pkg.apis.meta.v1.generated_pb2 import ObjectMeta

from modela.common import ObjectReference
from modela.ModelaException import ResourceNotFoundException, ResourceExistsException, GrpcErrorException


class Resource:
    """
    Base class for all classes representing Modela custom resources
    """
    DefaultVersion = 'v0.0.1'

    def __init__(self, resource: Message, client=None, name="", namespace="", version='v0.0.1'):
        self._object: Message = resource

        if client is not None:
            self._client = client

        if name != "" and namespace != "":
            if hasattr(self._client, "get"):
                try:
                    self._object = self._client.get(namespace, name).raw_message
                except ResourceNotFoundException:
                    self._object.metadata.CopyFrom(ObjectMeta())
                    self._object.metadata.name = name
                    self._object.metadata.namespace = namespace
                    self.default()
                    if hasattr(self._object.spec, 'versionName'):
                        self._object.spec.versionName = version

            else:
                print("Unable to lookup resource: object has no client repository")

    @property
    def spec(self):
        return self._object.spec

    @property
    def metadata(self) -> ObjectMeta:
        return self._object.metadata

    @property
    def name(self) -> str:
        return self.metadata.name

    @property
    def namespace(self) -> str:
        return self.metadata.namespace

    def has_label(self, label: str) -> bool:
        return label in self.metadata.labels

    def set_label(self, label: str, value: str):
        self.metadata.labels[label] = value

    def label(self, label: str) -> str:
        return self.metadata.labels[label]

    @property
    def raw_message(self) -> Message:
        return self._object

    @property
    def reference(self) -> ObjectReference:
        return ObjectReference(Namespace=self.namespace, Name=self.name)

    def submit(self, replace=False, **kwargs):
        """
        Submit resource to the cluster

        :param replace: Replace the resource if it already exists on the cluster.
        """
        self.ensure_client_repository()
        if replace:
            self._client.delete(self.namespace, self.name)
            name, namespace = self.name, self.namespace
            self._object.metadata.Clear()
            self._object.metadata.CopyFrom(ObjectMeta())
            self._object.metadata.name = name
            self._object.metadata.namespace = namespace

        try:
            self._client.create(self, **kwargs)
        except GrpcErrorException as ex:
            if not replace:
                raise ex

            time.sleep(0.5)  # Avoid race condition
            self._client.create(self, **kwargs)


    def update(self):
        """ Update the resource on the cluster with the latest data from the SDK """
        self.ensure_client_repository()
        self._client.update(self)

    def delete(self):
        """ Delete the resource on the cluster """
        self.ensure_client_repository()
        self._client.delete(self.namespace, self.name)
        name, namespace = self.name, self.namespace
        self._object.metadata.Clear()
        self._object.metadata.CopyFrom(ObjectMeta())
        self._object.metadata.name = name
        self._object.metadata.namespace = namespace

    def sync(self):
        """ Update the resource with the latest information from the cluster """
        self.ensure_client_repository()
        self._object = self._client.get(self.namespace, self.name).raw_message

    def ensure_client_repository(self):
        """ Assert that the client repository exists """
        assert hasattr(self, "_client")

    def default(self):
        print("Resource {0} is missing a default constructor; you may encounter errors upon creation.\n"
              "Please file an issue at https://github.com/metaprov/modela-python-sdk/issues".format(
            self.__class__.__name__))

    def __repr__(self):
        return "<{0} Resource at {1}/{2}>".format(self.__class__.__name__, self.namespace, self.name)

    @staticmethod
    def set_default_version(version):
        Resource.DefaultVersion = version



import grpc
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import ModelCompilerRun as MDModelCompilerRun
from github.com.metaprov.modelaapi.services.modelcompilerrun.v1.modelcompilerrun_pb2_grpc import ModelCompilerRunServiceStub
from github.com.metaprov.modelaapi.services.modelcompilerrun.v1.modelcompilerrun_pb2 import CreateModelCompilerRunRequest, \
    UpdateModelCompilerRunRequest, \
    DeleteModelCompilerRunRequest, GetModelCompilerRunRequest, ListModelCompilerRunsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class ModelCompilerRun(Resource):
    def __init__(self, item: MDModelCompilerRun = MDModelCompilerRun(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class ModelCompilerRunClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: ModelCompilerRunServiceStub = stub

    def create(self, modelcompilerrun: ModelCompilerRun) -> bool:
        request = CreateModelCompilerRunRequest()
        request.modelcompilerrun.CopyFrom(modelcompilerrun.raw_message)
        try:
            response = self.__stub.CreateModelCompilerRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, modelcompilerrun: ModelCompilerRun) -> bool:
        request = UpdateModelCompilerRunRequest()
        request.modelcompilerrun.CopyFrom(modelcompilerrun.raw_message)
        try:
            self.__stub.UpdateModelCompilerRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[ModelCompilerRun, bool]:
        request = GetModelCompilerRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetModelCompilerRun(request)
            return ModelCompilerRun(response.modelcompilerrun, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteModelCompilerRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteModelCompilerRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[ModelCompilerRun], bool]:
        request = ListModelCompilerRunsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListModelCompilerRuns(request)
            return [ModelCompilerRun(item, self) for item in response.modelcompilerruns.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





from abc import ABC
import modela.data.common as data_common
from modela.data.common import *
from modela.infra.Account import Account
from modela.infra.UserRoleClass import UserRoleClass
from modela.infra.models import Workload, NotificationSettings, OutputLogs, GitSettings, ImageLocation
from modela.training.common import *
from modela.common import *
from modela.Configuration import *
from dataclasses import field
from typing import List, Union, Dict
from github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 import CompilerSpec, Stakeholder, PermissionsSpec
import github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 as catalog_pb
import github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 as data_pb
import github.com.metaprov.modelaapi.services.common.v1.common_pb2 as common_pb
from modela.common import Plot
from modela.util import TrackedList


@datamodel(proto=data_pb.DataLocation)
class DataLocation(Configuration):
    """
    DataLocation describes the external location of data that will be accessed by Modela, and additional
    information on how to query the data if the location is a non flat-file source.
    """
    Type: DataLocationType = DataLocationType.ObjectStorage
    """
    The type of location where the data resides, which can either be an object inside an object storage system (i.e. Minio), a SQL location
    like a table or a view, a data stream (i.e. Kafka, currently unsupported), or a web location (currently unsupported)
    """
    ConnectionName: str = ''
    """
    In the case of the type of location being a database, ConnectionName specifies the name of the Connection resource
    that exists in the same tenant as the resource specifying the DataLocation. Modela will attempt to connect
    to the database using the credentials specified in the Connection, and will execute the query specified by the SQL field
    """
    BucketName: str = ''
    """
    In the case of the location type being an object storage system, BucketName is the name of the VirtualBucket resource
    that exists in the same tenant as the resource specifying the DataLocation. Modela will connect to the external
    object storage system, and will access the file from the path specified by the Path field
    """
    Path: str = ''
    """
    The path to a flat-file inside an object storage system. When using the Modela API to upload files (through the
    FileService API), Modela will upload the data to a predetermined path based on the Tenant, DataProduct,
    DataProductVersion, and resource type of the resource in relation to the file being uploaded.
    The path does not need to adhere to this format; you can still pass the path of a file inside a bucket not managed by Modela
    """
    Table: str = ''
    """ The name of a table inside a database, if applicable """
    Database: str = ''
    """ The name of a database inside the database system specified by the ConnectionName field """
    Sql: str = ''
    """ The SQL statement which will be executed to query data from the table specified by Table """
    Topic: str = ''
    """ The name of the streaming topic (currently unsupported) """


@datamodel(proto=data_pb.SampleSpec)
class SampleSettings(Configuration):
    """ SampleSpec specifies how the contents of a dataset should be sampled """
    Enabled: bool = False
    """ Indicates if sampling is enabled """
    Type: SamplingType = SamplingType.Random
    """ The type of sampling (random sampling, by default) """
    Rows: int = 500
    """ The number of rows to sample (by default, 500) """
    Percent: int = 100
    """ The percentage of rows to sample """
    Filter: str = ''
    """ The filter formula, valid only if the sample type is a filter """
    Column: str = ''
    """ The name of the column to be used for stratified sampling """


@datamodel(proto=data_pb.GovernanceSpec)
class GovernanceSpec(Configuration):
    """ GovernanceSpec describes the governance requirements for models produced under a DataProduct """
    Enabled: bool = False
    """ Indicates if governance is enabled """
    Country: str = ''
    """ The country whose regulations are under consideration """
    ItReviewer: str = ''
    """ The account name of the IT reviewer """
    ComplianceReviewer: str = ''
    """ The account name of the compliance reviewer """
    BusinessReviewer: str = ''
    """ The account name of the business reviewer """


@datamodel(proto=catalog_pb.CompilerSpec)
class CompilerSettings(Configuration):
    """ Compiler spec is used when there is a specification for model compilation """
    Enable: bool = False
    """ Enable set the enable to """
    Compiler: CompilerType = CompilerType.Nothing
    """
    Set one or more targets
    Enable set the enable to
    """
    Targets: List[HardwareTarget] = field(default_factory=lambda : [])
    """ Set one or more targets for the compiler """


ColorType = Color
ImageLocationType = ImageLocation
DataLocationType = DataLocation


@datamodel(proto=catalog_pb.Stakeholder)
class Stakeholder(Configuration):
    """ Stakeholder specifies the User Role Classes of an individual Account """
    Account: str = ''
    """ The name of an Account """
    Roles: List[ObjectReference] = field(default_factory=lambda : [])
    """ The object references to UserRoleClass resources which describe the actions the Account may perform """


@datamodel(proto=catalog_pb.PermissionsSpec)
class Permissions(Configuration):
    """
    PermissionsSpec specifies the Accounts that have access to a DataProduct or Tenant namespace and what permissions
    they possess for resources under the namespace
    """
    Stakeholders: List[Stakeholder] = field(default_factory=lambda : [])

    @classmethod
    def create(cls, accounts: Dict[Union[Account, str], Union[UserRoleClass, str, List[UserRoleClass], List[str]]], tenant='default-tenant') ->PermissionsSpec:
        """
        Generate a permission specification based on a dictionary of Accounts and their User Role Classes
        :param accounts: The dictionary of accounts to generate a permission specification from
        :param tenant: The name of the tenant which the accounts are based under. This parameter is not required if
          any User Role Class or Account resource is passed as an object (and not as a string)
        """
        stakeholders = []
        for account, roles in accounts.items():
            roleList, outList = roles if type(roles) == list else [roles], []
            for role in roleList:
                if type(account) == Account:
                    tenant = account.namespace
                    account = account.name
                if type(role) == UserRoleClass:
                    tenant = role.namespace
                    role = role.name
                outList.append(ObjectReference(Namespace=tenant, Name=role))
            stakeholders.append(Stakeholder(Account=account, Roles=outList))
        for stakeholder in stakeholders:
            for role in stakeholder.Roles:
                role.Namespace = tenant
        return cls(Stakeholders=stakeholders)


DataProductPermissions = Permissions


@datamodel(proto=data_pb.DataProductSpec)
class DataProductSpec(Configuration):
    """ DataProductSpec defines the desired state of the DataProduct """
    Owner: str = 'no-one'
    """ The name of the Account which created the object, which exists in the same tenant as the object """
    TenantRef: ObjectReference = ObjectReference(Namespace='modela-system', Name='default-tenant')
    """ The reference to the Tenant which owns the DataProduct. Defaults to `default-tenant` """
    GitLocation: GitSettings = GitSettings()
    """ GitLocation is the default Git location where all child resources will be tracked as YAML """
    ImageLocation: ImageLocationType = ImageLocation()
    """ ImageLocation is the default Docker image repository where model images produced under the DataProduct will be stored """
    LabName: str = 'default-lab'
    """ The name of the Lab that will be used by default with all compute-requiring child resources """
    ServingSiteName: str = 'default-serving-site'
    """ The name of the Serving Site which will be used by default with all Predictor resources """
    Task: TaskType = TaskType.BinaryClassification
    """ Task denote the machine learning task of the product (classification/regression,etc.) """
    Description: str = ''
    """ User-provided description of the object """
    DataLocation: DataLocationType = DataLocationType()
    """
    The default location for all artifacts created under the DataProduct. All data-producing resources will
    use the VirtualBucket specified by the DataLocation by default
    """
    Notification: NotificationSettings = NotificationSettings()
    """ The default notification specification for all resources under the DataProduct """
    TrainingResources: Workload = Workload('general-large')
    """ The default resource allocation for model training and data workloads that takes place under the DataProduct """
    ServingResources: Workload = Workload('general-large')
    """ The default resource allocation for model serving workloads that takes place under the DataProduct """
    RetriesOnFailure: int = 3
    """ Specifies how many times Jobs created under the DataProduct namespace will retry after failure """
    OnCallAccountName: str = ''
    """ The name of the Account which should be responsible for events that occur under the DataProduct """
    Compilation: CompilerSettings = CompilerSettings()
    """ The default compilation specification for Study resources created under the DataProduct """
    ClearanceLevel: SecurityClearanceLevel = SecurityClearanceLevel.Unclassified
    """
    The clearance level required to access the DataProduct. Accounts which do not have a clearance level
    greater than or equal to ClearanceLevel will be denied access to the DataProduct namespace
    """
    Priority: PriorityLevel = PriorityLevel.Medium
    """ The default priority level assigned to Jobs created under the DataProduct namespace """
    Color: ColorType = ColorType.NoColor
    """ The color assigned to the product, for visual purposes only """
    Governance: GovernanceSpec = GovernanceSpec()
    """ The Governance requirements (not functional as of the current release) """
    Public: bool = False
    """ Indicates if the DataProduct is public and can be accessed without permissions """
    Permissions: DataProductPermissions = DataProductPermissions()
    """
    Permissions denotes the specification that determines which Accounts
    can access the DataProduct and what actions they can perform
    """


@datamodel(proto=data_pb.DataProductVersionSpec)
class DataProductVersionSpec(Configuration):
    """ DataProductVersionSpec defines the desired state of a DataProductVersion """
    ProductRef: ObjectReference = ObjectReference(Namespace='default-tenant', Name='iris-product')
    """
    ProductRef contains the object reference to the DataProduct
    resource which the DataProductVersion describes the version of
    """
    Description: str = ''
    """ User-provided description of the object """
    PrevVersionName: str = ''
    """ The name of the version which preceded the current version """
    Baseline: bool = False
    """ Indicates if the version is a baseline, and if so will cause Modela to garbage collect the resources from previous versions """
    Owner: str = 'no-one'
    """ The name of the Account which created the object, which exists in the same tenant as the object """


@datamodel(proto=data_pb.CsvFileSpec)
class CsvFileFormat(Configuration):
    """ CsvFileSpec specifies the format of a CSV (comma-separated values) file """
    ColumnDelimiter: Delimiter = Delimiter.Comma
    """ The character used to separate fields (by default, a comma) """
    RowDelimiter: Delimiter = Delimiter.CRLF
    """ The character used to signal the end of a row (by default, a newline \\n) """
    EscapeChar: EscapeCharacter = EscapeCharacter.DoubleQuote
    """ The character used to escape the delimiter """
    Quote: QuoteCharacter = QuoteCharacter.DoubleQuote
    """ The charcter used for quotes (by default, a double quote ") """
    CommentChars: str = '#'
    """ The comment character used to split comments off the end of lines (by default, a hashtag #) """
    Header: bool = True
    """ Indicates if a header is present in the file """
    SkipRows: int = 0
    """ The number of rows to skip from the top of the file """
    NullValues: str = ''
    """ NullValues is a sequence of values to replace with NA. """
    Encoding: FileEncoding = FileEncoding.Utf8
    """ The unicode encoding of the file (e.g. 'utf-8' for UTF-8 encoded text) """
    MaxRows: int = 0
    """ The maximum number of rows to read """
    Strict: bool = True
    """ Indicates if the read of the CSV file should fail if there are any errors """
    Compression: CompressionType = CompressionType.Uncompressed
    """ The compression type, if the file is compressed """
    HasIndexColumn: bool = False
    """ Indicates if the file contains an index column """
    IndexColumn: int = 0
    """ The position of the index column """


@datamodel(proto=data_pb.ExcelSheetArea)
class ExcelSheetArea(Configuration):
    """ ExcelSheetArea specifies the bounds of the data within an excel sheet """
    EntireSheet: bool = True
    """
    Indicates if the excel reader should read the entire sheet; if false, it will read only within the bounds
    specified by the `To` and `From` fields of the ExcelSheetArea
    """
    FromColumn: int = 1
    """ If reading part of the excel sheet, start with the column in this position """
    ToColumn: int = 1
    """ If reading part of the excel sheet, end with the column in this position """
    FromRow: int = 1
    """ If reading part of the excel sheet, start with the row in this position """
    ToRow: int = 1
    """ If reading part of the excel sheet, end with the row in this position """


@datamodel(proto=data_pb.ExcelNotebookSpec)
class ExcelNotebookFormat(Configuration):
    """ ExcelNotebookSpec specifies the format of an excel file """
    FirstSheetWithData: bool = False
    """ Indicates if the excel reader should use the first sheet that contains data """
    SheetName: str = ''
    """ The name of the sheet that exists in the excel file to read data from """
    SheetIndex: int = 0
    """ The index of the sheet in the excel file to read data from """
    ColumnNameRow: int = 1
    """ The position of the row that contains the column names (i.e. the header) """
    Data: ExcelSheetArea = ExcelSheetArea()
    """ The specification for the bounds of the data """


@datamodel(proto=data_pb.Column)
class Column(Configuration):
    """
    Column specifies the attribute of a single column in a dataset. The fields of the Column align with the
    JSON schema standard; you can view detailed documentation at https://json-schema.org/draft/2020-12/json-schema-validation.html
    """
    Name: str = ''
    """ The name of the column """
    Datatype: DataType = DataType.Categorical
    """ The data type of the feature (e.g. number, string, boolean, etc.) """
    Format: DataDomain = DataDomain.Unknown
    """ The data domain of the feature, which constrains the contents of the feature to a specific set of values """
    DisplayName: str = ''
    """ The display name of the column, which is used in reports and other visual elements. If omitted, it will use the raw name """
    Description: str = ''
    """ The user-specified description of the feature """
    Ignore: bool = False
    """ Indicates if the feature should be ignored when building models """
    Target: bool = False
    """ Indicates if the feature is the target feature of the model, and the feature which predictions will be made on """
    Nullable: bool = False
    """ Indicates if the column can contain null values """
    Pk: bool = False
    """ Denotes if the column specifies a primary key of a database table (i.e. a users ID) """
    Fk: bool = False
    """ Denotes if the column specifies a foreign key of another database table """
    MultipleOf: int = 0
    """ The integer value which the values of the column should be a multiple of """
    Maximum: int = 0
    """ The maximum value of values all values in the column """
    ExclusiveMinimum: bool = False
    """ The exclusive lower limit of all values in the column, which does not include the minimum value """
    ExclusiveMaximum: bool = False
    """ The exclusive upper limit of all values in the column, which does not include the maximum value """
    Minimum: int = 0
    """ The minimum value of values all values in the column """
    MaxLength: int = 0
    """ The maximum length of values in the column, if the column data type is a string """
    MinLength: int = 0
    """ The minimum length of values in the column, if the column data type is a string """
    Pattern: str = ''
    """ The regex pattern which values in the column must adhere to """
    Required: bool = False
    """ Required """
    Example: str = ''
    """ A user-specified example value """
    ExternalDocs: str = ''
    """ A link to user-specified external documentation """
    Enum: List[str] = field(default_factory=lambda : [])
    """ The collection of unique values for categorical features """
    Ordinal: bool = False
    """ Indicates if the feature is ordinal, in the case of categorical features """
    MaxItems: int = 0
    """ The maximum number of items if the column is a list of values """
    MinItems: int = 0
    """ The minimum number of items if the column is a list of values """
    UniqueItems: bool = False
    """ Enforce that all the items in the list are unique """
    TimeColumn: bool = False
    """
    Indicates if the column is used as the time axis in time series forecasting
    Default is false.
    """
    Pii: bool = False
    """ Indicates if the column contains personally identifiable information """
    Phi: bool = False
    """ Indicates if the column contains personal health information """
    PersonalData: bool = False
    """ Indicates if the column contains any personal data """
    Protected: bool = False
    """ Protected means that this feature is important for ethical AI / Fairness """
    DefaultValueNum: float = 0
    """ The default value for number types; used internally for synthetic data and validation """
    Log: bool = False
    """ Indicates if values from this column will be sampled on a logarithmic scale """
    Mu: float = 0
    """ Mu is the mean of the normal distribution """
    Sigma: float = 0
    """ Sigma is the standard deviation of the distribution """
    SkewThreshold: float = None
    """ The threshold skew for skew detection """
    DriftThreshold: float = None
    """ The threshold drift value for model drift detection. """
    Index: bool = True
    """ Indicates if the column is an index column """
    Fold: bool = False
    """ Indicates if the column holds fold values """
    Weight: bool = False
    """ If True than this is a weight column """
    Reserved: bool = False
    """ Indicates that the feature should always be used in training """
    Imputation: Imputation = Imputation.AutoImputer
    """ The recommended imputation method for the column """
    Scaling: Scaling = Scaling.Auto
    """ The recommended scaling method for the column """
    Generated: bool = False
    """ Indicates if the feature was automatically generated """
    Formula: str = ''
    """ The formula used to generate the column """
    Id: bool = False
    """ Indicates if the column is an ID column """
    Step: float = 1
    """ The step value if the column values are a sequence of numbers """


@datamodel(proto=data_pb.TimeSeriesSchema)
class TimeSeriesSchema(Configuration):
    Freq: Frequency = None
    Country: HolidayCountry = None
    """ The holiday which should be taken into account """


@datamodel(proto=data_pb.RecommendationSchema)
class RecommendationSchema(Configuration):
    UserIDColumn: str = 'user_id'
    ItemIDColumn: str = 'item_id'
    """ The name of the column that specifies item IDs """
    RatingColumn: str = 'rating'
    """ The name of the column that specifies ratings """


@datamodel(proto=data_pb.MultiDatasetValidation)
class MultiDatasetValidation(Configuration):
    """ MultiDatasetValidation specifies a validation rule between multiple datasets """
    Type: MultiDatasetValidationRule = None
    """ The type of validation rule """
    Datasets: List[str] = field(default_factory=lambda : [])
    """ The list of datasets that will be checked """
    Generated: bool = False
    """ Indicates if this validation rule was automatically generated by the system """


@datamodel(proto=data_pb.DatasetValidation)
class DatasetValidation(Configuration):
    """ DatasetValidation specifies a validation rule for the contents of an entire dataset """
    Type: DatasetValidationRule = None
    """ The type of validation rule """
    Value: float = 0
    Min: float = 0
    Max: float = 0
    ValueSet: List[str] = field(default_factory=lambda : [])
    StrictMin: bool = False
    StrictMax: bool = False
    Generated: bool = False
    """ Indicates if this validation rule was automatically generated by the system """


@datamodel(proto=data_pb.MultiColumnValidation)
class MultiColumnValidation(Configuration):
    """ MultiColumnValidation specifies a validation rule that encompasses multiple columns """
    Type: MultiColumnValidationRule = None
    """ The type of validation rule """
    Columns: List[str] = field(default_factory=lambda : [])
    """ The list of columns that will be checked """
    Value: float = 0
    Min: float = 0
    Max: float = 0
    ValueSet: List[str] = field(default_factory=lambda : [])
    StrictMin: bool = False
    StrictMax: bool = False
    Generated: bool = False
    """ Indicates if this validation rule was automatically generated by the system """


@datamodel(proto=data_pb.ColumnValidation)
class ColumnValidation(Configuration):
    """ ColumnValidation specifies a validation rule for a single column """
    Type: ColumnValidationRule = None
    """ The type of validation rule """
    Column: str = ''
    """ The name of the column to be validated """
    Value: float = 0
    Min: float = 0
    Max: float = 0
    ValueSet: List[str] = field(default_factory=lambda : [])
    StrictMin: bool = False
    StrictMax: bool = False
    Generated: bool = False
    """ Indicates if this validation rule was automatically generated by the system """


@datamodel(proto=data_pb.FileValidation)
class FileValidation(Configuration):
    Type: FileValidationRule = None
    BucketName: str = ''
    Column: str = ''
    Value: float = 0
    """ For validation that need a single value """
    Min: float = 0
    Max: float = 0
    ValueSet: List[str] = field(default_factory=lambda : [])
    StrictMin: bool = False
    StrictMax: bool = False
    Generated: bool = False
    """ Generated indicate weather this validation rule was automatically generated """


@datamodel(proto=data_pb.ValidationSpec)
class ValidationRules(Configuration):
    MultiDatasetValidations: List[MultiDatasetValidation] = field(default_factory=lambda : [])
    DatasetValidations: List[DatasetValidation] = field(default_factory=lambda : [])
    """ DatasetValidations contains validations for the whole dataset """
    MultiColumnValidations: List[MultiColumnValidation] = field(default_factory=lambda : [])
    """ MultiColumnValidations defines validations for multiple columns from the dataset """
    ColumnValidations: List[ColumnValidation] = field(default_factory=lambda : [])
    """ ColumnValidations defines assertions for columns from the dataset """
    FileValidations: List[FileValidation] = field(default_factory=lambda : [])
    """ FileValidations defines assertions for the contents of the data file """


MDTimeSeriesSchema = TimeSeriesSchema
MDRecommendationSchema = RecommendationSchema


@datamodel(proto=data_pb.Schema)
class Schema(Configuration):
    """ Schema defines the column-level format and validation rules for data associated with a DataSource """
    Columns: List[Column] = field(default_factory=lambda : [])
    """ The collection of columns and their attributes """
    TimeSeriesSchema: MDTimeSeriesSchema = None
    """ The time-series schema, which sets time-series specific parameters """
    RecommendationSchema: MDRecommendationSchema = None
    """ The recommendation schema, which is used for the recommendation ML task """
    Validation: ValidationRules = ValidationRules()
    """ The specification for validation rules which will be performed on new Datasets """


@datamodel(proto=data_pb.RelationshipSpec)
class ColumnRelationship(Configuration):
    """ RelationSpec defines a relationship between two DataSource objects """
    Type: str = None
    """ The name of the relationship """
    Column: str = None
    """ The name of the columns that holds the foreign key """
    Arity: RelationshipArity = None
    """ The relationship arity """
    RelatesTo: str = None
    """ The name of the other DataSource object """


@datamodel(proto=data_pb.CorrelationSpec)
class CorrelationSetting(Configuration):
    """ CorrelationSpec specifies how the correlations between features in a Dataset should be computed """
    Cutoff: float = 50
    """ The minimum value of a computed correlation to be stored as a result """
    Method: str = 'pearson'
    """ The method to be used when computing correlations """
    Top: int = 10
    """ The number of top correlations to be included in the correlation results """


DataSourceSchema = Schema


@datamodel(proto=data_pb.DataSourceSpec)
class DataSourceSpec(Configuration):
    """ DataSourceSpec defines the desired state of the DataSource """
    Schema: DataSourceSchema = DataSourceSchema()
    """ The schema which will be used during the ingestion process of any Dataset resources which specify the DataSource """
    Sample: SampleSettings = SampleSettings()
    """
    The specification for how incoming data should be sampled (i.e. how many rows should be used). Applicable
    primarily for very large datasets
    """
    Relationships: List[ColumnRelationship] = field(default_factory=lambda : [])
    """ List of relationships to other data sources """
    Owner: str = 'no-one'
    """ The name of the Account which created the object, which exists in the same tenant as the object """
    VersionName: str = 'v0.0.1'
    """
    The name of the DataProductVersion which describes the version of the resource
    that exists in the same DataProduct namespace as the resource
    """
    Description: str = ''
    """ User-provided description of the object """
    FileType: FlatFileType = FlatFileType.Csv
    """ The file type of incoming data which uses the DataSource (by default, a CSV file) """
    Task: TaskType = TaskType.BinaryClassification
    """
    The default task for Dataset resources created from the DataSource. If null, this will be determined from
    the default task of the DataProduct which owns the DataSource
    """
    Csvfile: CsvFileFormat = CsvFileFormat()
    """ The file format for CSV files, if applicable """
    ExcelNotebook: ExcelNotebookFormat = None
    """ The file format of excel files, if applicable """
    DatasetType: DatasetType = DatasetType.Tabular
    """ The type of dataset; currently, the only supported type is `tabular` """


@datamodel(proto=data_pb.DatasetSpec)
class DatasetSpec(Configuration):
    """ DatasetSpec defines the desired state of the Dataset """
    Origin: DataLocation = DataLocation()
    """
    Origin is the location of the data file or database query which holds the raw data of the Dataset. When the Dataset is
    created, the resource controller will retrieve the data from the location, validate it against its Data Source
    if applicable, and store it inside the `live` section of the Virtual Bucket resource specified by the location
    """
    LabRef: ObjectReference = ObjectReference('default-tenant', 'default-lab')
    """ The reference to the Lab under which Jobs created by the Dataset will be executed """
    Location: DataLocation = DataLocation()
    """
    Location is the final location of the data which was copied from the `Origin` location during the ingestion phase.
    This field is set by the Dataset resource controller and should not be changed by any end-users
    """
    Owner: str = 'no-one'
    """ The name of the Account which created the object, which exists in the same tenant as the object """
    VersionName: str = 'v0.0.1'
    """
    The name of the DataProductVersion which describes the version of the resource
    that exists in the same DataProduct namespace as the resource
    """
    DatasourceName: str = ''
    """
    The reference to the Data Source resource which exists in the same Data Product namespace as the object.
    The Data Source must represent the columns and the task type of the Dataset. The validation rules associated with
    the Data Source will be validated against the raw data of the Dataset once it is created
    """
    Description: str = ''
    """ User-provided description of the object """
    DisplayName: str = ''
    """ User-provided display name of the object """
    Reported: bool = True
    """ Indicates if a PDF report containing the Dataset's profile should be generated """
    Snapshotted: bool = False
    """
    Indicates if the resource controller has created a snapshot of the data in the case that it is being read
    directly from a database, and must be converted to a flat-file type such as a CSV as a result
    """
    Validate: bool = True
    """ Indicates if the Dataset should be checked against the validation rules of its Data Source """
    Labeled: bool = True
    """ Indicates if the data is labeled or not """
    Synthetic: bool = False
    """ Indicates if synthetic data should be generated (currently unimplemented) """
    SyntheticRows: int = 0
    """ If `Synthetic` is set to true, SyntheticRows indicates how many rows of synthetic data should be generated """
    Resources: Workload = None
    """ Resources specifies the resource requirements which the Dataset will request when creating Jobs to analyze the data """
    ActiveDeadlineSeconds: int = 600
    """ The deadline in seconds for all Jobs created by the Dataset """
    Type: DatasetType = DatasetType.Tabular
    """ The type of dataset which was uploaded. `tabular` is the only supported type as of the current release """
    Sample: SampleSettings = None
    """
    The specification for how the data should be sampled, if applicable. Sampling may improve dataset and model creation
    time in the case of very large datasets that are being rapidly prototyped and iterated on
    """
    Task: TaskType = None
    """ The machine learning task relevant to the Dataset. This field *must* be the same as the Data Source of the object """
    Notification: NotificationSettings = None
    """ The notification specification that determines which notifiers will receive Alerts generated by the object """
    Correlation: CorrelationSetting = None
    """
    The specification for how to find the correlations of the Dataset's features during the profiling phase.
    Based on the specification, the data plane will compute the correlation between each feature and will store the highest-scoring
    """
    Fast: bool = False
    """
    Indicates if the Dataset should be quickly processed.
    If enabled, the validation, profiling, and reporting phases will be skipped.
    """


@datamodel(proto=data_pb.ColumnStatistics)
class ColumnStatistics(Configuration):
    """ ColumnStatistics contains statistical parameters for a single feature from a dataset """
    Name: str = ''
    """ The name of the column """
    Datatype: DataType = None
    """ The data type of the column """
    Count: float = 0
    """ Amount of rows which contain a value for the feature """
    Distinct: int = 0
    """ Amount of unique values present in the column """
    Missing: int = 0
    """ Amount of missing values present in the column """
    PercentMissing: float = 0
    """ Percentage of missing values in the column """
    Mean: float = 0
    """ The mean of all values in the column, if the column data type is a number """
    Stddev: float = 0
    """ The standard deviation of the columns values """
    Variance: float = 0
    """ The variability of the columns values from the columns mean """
    Min: float = 0
    """ The minimum value of all values in the column """
    Max: float = 0
    """ The maximum value of all values in the column """
    Kurtosis: float = 0
    """ The computed kurtosis, which measures the peakedness of the distribution of values in the column """
    Skewness: float = 0
    """ The computed skewness, which measures the asymmetry of the distribution of values in the column """
    Sum: float = 0
    """ Skewness is the standard deviation value of the attribute """
    Mad: float = 0
    """ The sum of all values in the column """
    P25: float = 0
    """ The 25% point of all the values of the column in order """
    P50: float = 0
    """ The 50% point of all the values of the column in order, also known as the median """
    P75: float = 0
    """ The 75% point of all the values of the column in order """
    Iqr: float = 0
    """ The interquartile range of the columns values """
    Mode: str = ''
    """ The mode value of the column, also known as the most frequent value """
    Zeros: float = 0
    """ The number of zero values in the column """
    Invalid: int = 0
    """ The number of invalid values in the column """
    Importance: float = 0
    """ The feature importance of the column """
    Target: bool = False
    """ Indicates if the feature is the target attribute for a Study, as specified by the Dataset's DataSource """
    Ignore: bool = False
    """ Indicates if the column should be ignored, as specified by the Dataset's DataSource """
    Nullable: bool = False
    """ Indicates if the column may contain null values, as specified by the Dataset's DataSource """
    HighCardinality: bool = False
    """ Indicates if the column has high cardinality and should use the high cardinality encoder during feature engineering """
    HighCorrWithOtherFeatures: bool = False
    """ Indicates if the column has high correlation with another feature, and that it should be dropped """
    LowCorrWithTarget: bool = False
    """ Indicate that the feature has low correlation with the target feature, and that it should be dropped """
    HighMissingPct: bool = False
    """ Indicates if the column has a high percentage of missing values, and that it should be dropped """
    Skewed: bool = False
    """
    Marks that the column is skewed and would require a power transform.
    
    If skewness is less than -1 or greater than 1, the distribution is highly skewed.
    If skewness is between -1 and -0.5 or between 0.5 and 1, the distribution is moderately skewed.
    If skewness is between -0.5 and 0.5, the distribution is approximately symmetric
    """
    Id: bool = False
    """ Indicates if the column is an ID column, such as a primary key """
    Constant: bool = False
    Duplicate: bool = False
    """ Indicates if the column is a duplicate of another column """
    Reserved: bool = False
    """ Indicates if the column is reserved and must be a feature included in model training """
    Outliers: int = 0
    """ The amount of outliers detected in the columns values """
    Completeness: float = 0
    """ The ratio between non-null and null values in the column """
    DistinctValueCount: float = 0
    """ The ratio between unique values and non-unique values in the column """
    MostFreqValuesRatio: float = 0
    """ The ratio between most the most frequent value to the number of total values in the column """
    IndexOfPeculiarity: float = 0
    """ Used for text attributes """
    CorrToTarget: float = 0
    """ Correlation to the target feature """


@datamodel(proto=common_pb.ColumnProfile)
class ColumnProfile(Configuration):
    Name: str = ''
    Count: int = 0
    Type: str = 0
    Missing: int = 0
    PercentMissing: float = 0
    Distinct: int = 0
    Mean: float = 0
    Mode: str = ''
    Stddev: float = 0
    Variance: float = 0
    Min: float = 0
    Max: float = 0
    Kurtosis: float = 0
    Skewness: float = 0
    Sum: float = 0
    Mad: float = 0
    Zeros: float = 0
    P25: float = 0
    P50: float = 0
    P75: float = 0
    P100: float = 0
    Iqr: float = 0
    Cv: float = 0
    Top: int = 0
    Freq: int = 0
    Ignore: bool = False
    Target: bool = False
    Invalid: int = 0
    Importance: float = 0
    Nullable: bool = False
    HighCardinality: bool = False
    HighCorrWithOtherFeatures: bool = False
    LowCorrWithTarget: bool = False
    HighMissingPct: bool = False
    Skewed: bool = False
    Id: bool = False
    Constant: bool = False
    Duplicate: bool = False
    Reserved: bool = False
    Outliers: int = 0
    Completeness: float = 0
    DistinctValueCount: float = 0
    MostFreqValuesRatio: float = 0
    IndexOfPeculiarity: float = 0
    Values: List[str] = field(default_factory=lambda : [])
    CorrToTarget: float = 0


@datamodel(proto=common_pb.DatasetProfile)
class DatasetProfile(Configuration):
    Cols: int = 0
    Rows: int = 0
    Filesize: int = 0
    Imbalance: bool = False
    Plots: List[Plot] = field(default_factory=lambda : [])
    Columns: List[ColumnProfile] = field(default_factory=lambda : [])
    Hash: str = ''


@datamodel(proto=data_pb.Correlation)
class Correlation(Configuration):
    """ Correlation records the correlation between two features in a Dataset """
    Feature1: str = ''
    """ The first feature name """
    Feature2: str = ''
    """ The second feature name """
    Value: float = 0
    """ The correlation value """
    Method: str = ''
    """ How the value was calculated """


@datamodel(proto=data_pb.DatasetStatistics)
class DatasetStatistics(Configuration):
    """
    DatasetStatistics contains statistics about the Dataset's overall data, as well as every feature of the data. The
    data structure is populated with information during the `Profiling` phase of the parent Dataset.
    """
    Columns: List[ColumnStatistics] = field(default_factory=lambda : [])
    """ Columns contains the collection of statistics for each feature """
    Rows: int = 0
    """ Number of rows observed from the data """
    Cols: int = 0
    """ Number of columns observed from the data """
    FileSize: int = 0
    """ The file size of the data in bytes """
    CorrelationsWithTarget: List[Correlation] = field(default_factory=lambda : [])
    """ The top correlations between all features and the target feature """
    TopCorrelations: List[Correlation] = field(default_factory=lambda : [])
    """ The top correlations between features, computed per the CorrelationSpec of the parent Dataset """


@datamodel(proto=data_pb.DatasetCondition)
class DatasetCondition(Configuration):
    """ DatasetCondition describes the state of a dataset at a certain point """
    Type: data_common.DatasetCondition = data_common.DatasetCondition.Ready
    Status: ConditionStatus = ConditionStatus.ConditionUnknown
    """ Status of the condition, one of True, False, Unknown """
    LastTransitionTime: Time = None
    """ Last time the condition transitioned from one status to another. """
    Reason: str = ''
    """ The reason for the condition's last transition. """
    Message: str = ''
    """ A human readable message indicating details about the transition. """


@datamodel(proto=data_pb.DataValidationResult)
class DataValidationResult(Configuration):
    Type: str = ''
    Column: str = ''
    Error: str = ''
    Passed: bool = False


@datamodel(proto=data_pb.DatasetStatus)
class DatasetStatus(ImmutableConfiguration):
    """ DatasetStatus defines the observed state of a Dataset object """
    Statistics: DatasetStatistics = None
    """ Statistics for each column of the Dataset, which are generated during the profiling phase. """
    Phase: DatasetPhase = DatasetPhase.Pending
    """ The current phase of the Dataset progress """
    ReportName: str = ''
    """
    Reference to the report object that was generated for the dataset, which exists in the same Data Product namespace
    as the object
    """
    ReportUri: str = ''
    """ The location of report generated during the reporting phase. This field is intended for internal use """
    ProfileUri: str = ''
    """ The location of raw profile data. This field is intended for internal use """
    Imbalanced: bool = False
    """ Whether or not the data was detected as imbalanced """
    ObservedGeneration: int = 0
    """ ObservedGeneration is the last generation that was acted on """
    ValidationResults: List[DataValidationResult] = field(default_factory=lambda : [])
    """ List of validation results which are generated for every validation rule associated with the Dataset's Data Source """
    LastStudyTime: Time = None
    """ Last time the Dataset was used with a Study """
    FailureReason: StatusError = None
    """ In the case of failure, the Dataset resource controller will set this field with a failure reason """
    FailureMessage: str = ''
    """ In the case of failure, the Dataset resource controller will set this field with a failure message """
    Progress: int = 0
    """ The current progress of the Dataset, with a maximum of 100, that is associated with the current phase """
    Hash: str = ''
    """ Sha256 signature of the raw data. Intended for internal use """
    Logs: OutputLogs = OutputLogs()
    """ The log file specification that determines the location of all logs produced by the object """
    DerivedFromDataset: str = ''
    """ If the dataset is derived, the name of the Dataset that the object is derived from """
    LastUpdated: Time = None
    """ The last time the object was updated """
    StartTime: Time = None
    """ The time that the system started processing the Dataset, usually after the creation of the object """
    EndTime: Time = None
    """ The time that the Dataset finished processing, either due to completion or failure """
    Conditions: List[DatasetCondition] = field(default_factory=lambda : [])



import grpc
from github.com.metaprov.modelaapi.pkg.apis.infra.v1alpha1.generated_pb2 import Tenant as MDTenant
from github.com.metaprov.modelaapi.services.tenant.v1.tenant_pb2_grpc import TenantServiceStub
from github.com.metaprov.modelaapi.services.tenant.v1.tenant_pb2 import CreateTenantRequest, \
    UpdateTenantRequest, \
    DeleteTenantRequest, GetTenantRequest, ListTenantsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class Tenant(Resource):
    def __init__(self, item: MDTenant = MDTenant(), client=None, namespace="", name=""):
        super().__init__(item, client, namespace=namespace, name=name)


class TenantClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: TenantServiceStub = stub

    def create(self, tenant: Tenant) -> bool:
        request = CreateTenantRequest()
        request.tenant.CopyFrom(tenant.raw_message)
        try:
            response = self.__stub.CreateTenant(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, tenant: Tenant) -> bool:
        request = UpdateTenantRequest()
        request.tenant.CopyFrom(tenant.raw_message)
        try:
            self.__stub.UpdateTenant(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[Tenant, bool]:
        request = GetTenantRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetTenant(request)
            return Tenant(response.tenant, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteTenantRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteTenant(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[Tenant], bool]:
        request = ListTenantsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListTenants(request)
            return [Tenant(item, self) for item in response.tenants.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.team.v1alpha1.generated_pb2 import RunBook as MDRunBook
from github.com.metaprov.modelaapi.services.runbook.v1.runbook_pb2_grpc import RunBookServiceStub
from github.com.metaprov.modelaapi.services.runbook.v1.runbook_pb2 import CreateRunBookRequest, \
    UpdateRunBookRequest, \
    DeleteRunBookRequest, GetRunBookRequest, ListRunBooksRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class RunBook(Resource):
    def __init__(self, item: MDRunBook = MDRunBook(), client=None, namespace="", name=""):
        super().__init__(item, client, namespace=namespace, name=name)


class RunBookClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: RunBookServiceStub = stub

    def create(self, runbook: RunBook) -> bool:
        request = CreateRunBookRequest()
        request.runbook.CopyFrom(runbook.raw_message)
        try:
            response = self.__stub.CreateRunBook(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, runbook: RunBook) -> bool:
        request = UpdateRunBookRequest()
        request.runbook.CopyFrom(runbook.raw_message)
        try:
            self.__stub.UpdateRunBook(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[RunBook, bool]:
        request = GetRunBookRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetRunBook(request)
            return RunBook(response.runbook, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteRunBookRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteRunBook(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[RunBook], bool]:
        request = ListRunBooksRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListRunBooks(request)
            return [RunBook(item, self) for item in response.runbooks.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import DataPipelineRun as MDDataPipelineRun
from github.com.metaprov.modelaapi.services.datapipelinerun.v1.datapipelinerun_pb2_grpc import DataPipelineRunServiceStub
from github.com.metaprov.modelaapi.services.datapipelinerun.v1.datapipelinerun_pb2 import CreateDataPipelineRunRequest, \
    UpdateDataPipelineRunRequest, \
    DeleteDataPipelineRunRequest, GetDataPipelineRunRequest, ListDataPipelineRunRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class DataPipelineRun(Resource):
    def __init__(self, item: MDDataPipelineRun = MDDataPipelineRun(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class DataPipelineRunClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: DataPipelineRunServiceStub = stub

    def create(self, datapipelinerun: DataPipelineRun) -> bool:
        request = CreateDataPipelineRunRequest()
        request.datapipelinerun.CopyFrom(datapipelinerun.raw_message)
        try:
            response = self.__stub.CreateDataPipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, datapipelinerun: DataPipelineRun) -> bool:
        request = UpdateDataPipelineRunRequest()
        request.datapipelinerun.CopyFrom(datapipelinerun.raw_message)
        try:
            self.__stub.UpdateDataPipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[DataPipelineRun, bool]:
        request = GetDataPipelineRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetDataPipelineRun(request)
            return DataPipelineRun(response.datapipelinerun, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteDataPipelineRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteDataPipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[DataPipelineRun], bool]:
        request = ListDataPipelineRunRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListDataPipelineRuns(request)
            return [DataPipelineRun(item, self) for item in response.datapipelineruns.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import hashlib

import grpc
from github.com.metaprov.modelaapi.services.fileservices.v1.fileservices_pb2 import DataBlock
from github.com.metaprov.modelaapi.services.fileservices.v1.fileservices_pb2_grpc import FileServicesServiceStub

from modela.data.models import DataLocation
from modela.ModelaException import ModelaException
from tqdm import *

class DataBlockRequestIterable(object):
    BLOCK_SIZE = 200000

    def __init__(self, name: str,
                 data: bytes,
                 tenant: str,
                 data_product: str,
                 version: str,
                 bucket: str,
                 resource_type: str,
                 resource_name: str):

        self.name = name
        self.data = data
        self.tenant = tenant
        self.data_product = data_product
        self.version = version
        self.bucket = bucket
        self.resource_type = resource_type
        self.resource_name = resource_name
        self.loc = 0
        self.pbar = tqdm(total=len(data), desc="Uploading", unit="bytes", ncols=100)

    def __iter__(self):
        return self

    def __next__(self):
        data_block = self.data[self.loc:self.loc + self.BLOCK_SIZE]
        if len(data_block) > 0:
            hash = hashlib.new('md5', (data_block if type(data_block) == bytes else str(data_block).encode('utf-8'))).hexdigest()
            if type(data_block) == str:
                data_block = bytes(data_block, encoding='utf-8')
            request = DataBlock(
                name=self.name,
                data=data_block,
                md5_hash=str(hash),
                tenant=self.tenant,
                dataProductName=self.data_product,
                dataProductVersion=self.version,
                bucket=self.bucket,
                resourceType=self.resource_type,
                resourceName=self.resource_name
            )
            self.loc += self.BLOCK_SIZE
            self.pbar.update(min(self.BLOCK_SIZE, len(self.data)))
            return request
        else:
            raise StopIteration


class FileService:
    def __init__(self, stub):
        self.__stub: FileServicesServiceStub = stub

    def upload_file(self,
                    name: str,
                    data: bytes,
                    tenant: str,
                    data_product: str,
                    version: str,
                    bucket: str,
                    resource_type: str,
                    resource_name: str) -> DataLocation:
        data_block_iterable = DataBlockRequestIterable(name, data, tenant, data_product,
                                                       version, bucket, resource_type, resource_name)
        try:
            response = self.__stub.UploadChunk(data_block_iterable, timeout=20)
            return DataLocation(BucketName=bucket, Path=response.key)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)



from dataclasses import dataclass, field
from typing import List

from modela.Configuration import Configuration
from modela.common import StatusError, Time, ConditionStatus, SecretReference, ObjectReference
from modela.infra.common import ProviderName, ConnectionConditionType, ConnectionCategory
dataclass()


@dataclass
class AmazonAthenaConnection(Configuration):
    HostName: str = ""
    Username: str = ""
    Password: str = ""
    Region: str = ""
    Url: str = ""


@dataclass
class GcpBigQueryConnection(Configuration):
    Host: str = ""
    Port: int = 9042
    Username: str = ""
    Password: str = ""
    ProjectID: str = ""
    OauthType: str = ""
    ServiceAccountKeyPath: str = ""
    Url: str = ""


@dataclass
class ApacheCassandraConnection(Configuration):
    Host: str = ""
    Port: int = 9042
    Keyspace: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class ApacheDruidConnection(Configuration):
    Host: str = ""
    Port: int = 9042
    Keyspace: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class ApacheDrillConnection(Configuration):
    Host: str = ""
    Port: int = 9042
    Keyspace: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class AzureSQLDatabaseConnection(Configuration):
    Host: str = ""
    Port: int = 1433
    Username: str = ""
    Password: str = ""
    Database: str = ""
    Url: str = ""


@dataclass
class MongoDbConnection(Configuration):
    Host: str = ""
    Port: int = 27017
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class MySqlConnection(Configuration):
    Host: str = ""
    Port: int = 3306
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class OdbcConnection(Configuration):
    ConntectionString: str = ""


@dataclass
class OracleConnection(Configuration):
    Host: str = ""
    Port: int = 1521
    Sid: str = "XE"
    Driver: str = "thin"
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class PostgresSQLConnection(Configuration):
    Host: str = ""
    Port: int = 5432
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class PrestoConnection(Configuration):
    Host: str = ""
    Port: int = 8080
    Catalog: str = ""
    Schema: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class AmazonRedShiftConnection(Configuration):
    Host: str = ""
    Port: int = 5439
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class ApacheHiveConnection(Configuration):
    Host: str = ""
    Port: int = 10000
    Schema: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class SnowflakeConnection(Configuration):
    Host: str = ""
    Port: int = 443
    Username: str = ""
    Password: str = ""
    Database: str = ""
    Schema: str = ""
    Warehouse: str = ""
    Url: str = ""


@dataclass
class SybaseConnection(Configuration):
    Host: str = ""
    Port: int = 5000
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class VerticaConnection(Configuration):
    Host: str = ""
    Port: int = 5433
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class CockroachDBConnection(Configuration):
    Host: str = ""
    Port: int = 5432
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class ElasticSearchConnection(Configuration):
    Host: str = ""
    Port: int = 5432
    Prefix: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class InformixConnection(Configuration):
    Host: str = ""
    Port: int = 1526
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class SAPHanaConnection(Configuration):
    Host: str = ""
    Port: int = 1526
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class TeradataConnection(Configuration):
    Host: str = ""
    Port: int = 6666
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class GcpSpannerConnection(Configuration):
    Project: str = ""
    Instance: str = ""
    Username: str = ""
    Password: str = ""
    Database: str = ""
    Url: str = ""


@dataclass
class ApacheSparkConnection(Configuration):
    Host: str = ""
    Port: int = 5433
    Schema: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class MSSqlServerConnection(Configuration):
    Host: str = ""
    Port: int = 1433
    Database: str = ""
    Instance: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class ClickHouseConnection(Configuration):
    Host: str = ""
    Port: int = 8123
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class GreenPlumConnection(Configuration):
    Host: str = ""
    Port: int = 5432
    Database: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class CouchbaseConnection(Configuration):
    Host: str = ""
    Port: int = 8123
    Database: str = ""
    Username: str = ""
    Password: str = ""
    DefaultBucket: str = ""
    Url: str = ""


@dataclass
class ExasolConnection(Configuration):
    Host: str = ""
    Port: int = 0
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class SingleStoreConnection(Configuration):
    Host: str = ""
    Port: int = 0
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class Neo4JConnection(Configuration):
    Host: str = ""
    Port: int = 0
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class KafkaConnection(Configuration):
    Host: str = ""
    Port: int = 0
    Channel: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class SqliteConnection(Configuration):
    FileName: str = "general"


@dataclass
class AwsS3Connection(Configuration):
    DefaultRegion: str = ""
    AccessKey: str = ""
    SecretKey: str = ""


@dataclass
class AzureStorageConnection(Configuration):
    StorageAccount: str = ""
    AccessKey: str = ""


@dataclass
class GcpStorageConnection(Configuration):
    KeyFile: str = ""
    Project: str = ""


@dataclass
class DigitalOceanConnection(Configuration):
    Token: str = ""
    AccessKey: str = ""
    SecretKey: str = ""
    DefaultRegion: str = ""
    Host: str = ""


@dataclass
class MinioConnection(Configuration):
    AccessKey: str = ""
    SecretKey: str = ""
    Host: str = ""


@dataclass
class ImageRegistryConnection(Configuration):
    Host: str = ""
    Username: str = ""
    Password: str = ""


@dataclass
class SlackConnection(Configuration):
    WebhookUrl: str = ""
    Channel: str = ""
    UserName: str = ""
    Token: str = ""


@dataclass
class SmtpConnection(Configuration):
    Host: str = ""
    Port: int = 25
    UserName: str = ""
    Password: str = ""


@dataclass
class GithubConnection(Configuration):
    Token: str = ""
    UserName: str = ""
    Ssh: str = ""


@dataclass
class BitbucketConnection(Configuration):
    Token: str = ""
    UserName: str = ""
    Ssh: str = ""


@dataclass
class AliCloudConnection(Configuration):
    AccessKey: str = ""
    SecretKey: str = ""
    Region: str = ""
    Host: str = ""


@dataclass
class GitlabConnection(Configuration):
    Token: str = ""
    Ssh: str = ""


@dataclass
class HetznerConnection(Configuration):
    Token: str = ""
    Ssh: str = ""


@dataclass
class OpenstackConnection(Configuration):
    UserName: str = ""
    TenantName: str = ""
    Password: str = ""
    AuthUrl: str = ""
    Region: str = ""


@dataclass
class OvhConnection(Configuration):
    Endpoint: str = ""
    Application: str = ""
    Secret: str = ""
    Consumerkey: str = ""


@dataclass
class LinodeConnection(Configuration):
    Token: str = ""


@dataclass
class MSTeamConnection(Configuration):
    Webhook: str = ""


@dataclass
class MattermostConnection(Configuration):
    Url: str = ""
    Channel: str = ""
    Username: str = ""


@dataclass
class HipchatConnection(Configuration):
    Url: str = ""
    Token: str = ""
    Room: str = ""


@dataclass
class VictorOpConnection(Configuration):
    ApiID: str = ""
    ApiKey: str = ""
    Url: str = ""


@dataclass
class PagerDutyConnection(Configuration):
    ApiID: str = ""
    ApiKey: str = ""
    Url: str = ""


@dataclass
class PushoverConnection(Configuration):
    ApiID: str = ""
    ApiKey: str = ""
    Url: str = ""


@dataclass
class OpsgenieConnection(Configuration):
    ApiID: str = ""
    ApiKey: str = ""
    Url: str = ""


@dataclass
class WebhookConnection(Configuration):
    Url: str = ""


@dataclass
class GoogleSheetsConnection(Configuration):
    Scopes: str = ""
    Id: str = ""
    DataToPull: str = ""


@dataclass
class FTPConnection(Configuration):
    Host: str = ""
    Port: int = 9042
    Keyspace: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class SFTPConnection(Configuration):
    Host: str = ""
    Port: int = 9042
    Keyspace: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class DropboxConnection(Configuration):
    Token: str = ""


@dataclass
class BoxConnection(Configuration):
    Token: str = ""


@dataclass
class FacebookConnection(Configuration):
    Token: str = ""


@dataclass
class TwitterConnection(Configuration):
    Token: str = ""


@dataclass
class DB2Connection(Configuration):
    Host: str = ""
    Port: int = 9042
    Keyspace: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class DremioConnection(Configuration):
    Host: str = ""
    Port: int = 9042
    Keyspace: str = ""
    Username: str = ""
    Password: str = ""
    Url: str = ""


@dataclass
class FtpConnection(Configuration):
    Host: str = ""
    Port: int = 21
    Username: str = ""
    Password: str = ""


@dataclass
class SFtpConnection(Configuration):
    Host: str = ""
    Port: int = 21
    Username: str = ""
    Password: str = ""


@dataclass
class RabbitMQConnection(Configuration):
    Host: str = ""
    Port: int = 21
    Username: str = ""
    Password: str = ""


@dataclass
class Connection(Configuration):
    TenantRef: ObjectReference = None
    Category: ConnectionCategory = ConnectionCategory.General
    Provider: ProviderName = ProviderName.UnknownProvider
    SecretRef: SecretReference = None
    Athena: AmazonAthenaConnection = None
    Drill: ApacheDrillConnection = None
    Druid: ApacheDruidConnection = None
    Hive: ApacheHiveConnection = None
    Redshift: AmazonRedShiftConnection = None
    Dermio: DremioConnection = None
    Db2: DB2Connection = None
    Bigquery: GcpBigQueryConnection = None
    Cassandra: ApacheCassandraConnection = None
    Azuresqldatabase: AzureSQLDatabaseConnection = None
    Mongodb: MongoDbConnection = None
    Mysql: MySqlConnection = None
    Odbc: OdbcConnection = None
    Oracle: OracleConnection = None
    Postgres: PostgresSQLConnection = None
    Presto: PrestoConnection = None
    Snowflake: SnowflakeConnection = None
    Sybase: SybaseConnection = None
    Vertica: VerticaConnection = None
    Cockroachdb: CockroachDBConnection = None
    Elasticsearch: ElasticSearchConnection = None
    Informix: InformixConnection = None
    Hana: SAPHanaConnection = None
    Teradata: TeradataConnection = None
    Spanner: GcpSpannerConnection = None
    Spark: ApacheSparkConnection = None
    Mssqlserver: MSSqlServerConnection = None
    Clickhouse: ClickHouseConnection = None
    Greenplum: GreenPlumConnection = None
    Couchbase: CouchbaseConnection = None
    Exasol: ExasolConnection = None
    Sqlite: SqliteConnection = None
    Singlestore: SingleStoreConnection = None
    Gsheets: GoogleSheetsConnection = None
    Azurestorage: AzureStorageConnection = None
    Alicloud: AliCloudConnection = None
    Digitalocean: DigitalOceanConnection = None
    Hetzner: HetznerConnection = None
    Gcpstorage: GcpStorageConnection = None
    Linode: LinodeConnection = None
    Minio: MinioConnection = None
    Openstack: OpenstackConnection = None
    Ovh: OvhConnection = None
    Aws: AwsS3Connection = None
    Smtp: SmtpConnection = None
    Ftp: FtpConnection = None
    Sftp: SFtpConnection = None
    Dropbox: DropboxConnection = None
    Box: BoxConnection = None
    ImageRegistry: ImageRegistryConnection = None
    Github: GithubConnection = None
    Gitlab: GitlabConnection = None
    Bitbucket: BitbucketConnection = None
    Slack: SlackConnection = None
    Msteam: MSTeamConnection = None
    MatterMost: MattermostConnection = None
    Hipchat: HipchatConnection = None
    Victorop: VictorOpConnection = None
    Pagerduty: PagerDutyConnection = None
    Pushover: PushoverConnection = None
    Opsgenie: OpsgenieConnection = None
    Webhook: WebhookConnection = None
    Facebook: FacebookConnection = None
    Twitter: TwitterConnection = None
    Rabbitmq: RabbitMQConnection = None
    Kafka: KafkaConnection = None
    Neo4j: Neo4JConnection = None
    Owner: str = "no-one"

@dataclass
class ConnectionCondition(Configuration):
    Type: ConnectionConditionType = ConnectionConditionType.ConnectionReady
    Status: ConditionStatus = ConditionStatus.ConditionUnknown
    LastTransitionTime: Time = None
    Reason: str = ""
    Message: str = ""


@dataclass
class ConnectionStatus(Configuration):
    ObservedGeneration: int = 0
    LastUpdated: Time = None
    FailureReason: StatusError = None
    FailureMessage: str = ""
    Conditions: List[ConnectionCondition] = field(default_factory=lambda: [])




import grpc
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import ModelPipelineRun as MDModelPipelineRun
from github.com.metaprov.modelaapi.services.modelpipelinerun.v1.modelpipelinerun_pb2_grpc import ModelPipelineRunServiceStub
from github.com.metaprov.modelaapi.services.modelpipelinerun.v1.modelpipelinerun_pb2 import CreateModelPipelineRunRequest, \
    UpdateModelPipelineRunRequest, \
    DeleteModelPipelineRunRequest, GetModelPipelineRunRequest, ListModelPipelineRunsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class ModelPipelineRun(Resource):
    def __init__(self, item: MDModelPipelineRun = MDModelPipelineRun(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class ModelPipelineRunClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: ModelPipelineRunServiceStub = stub

    def create(self, modelpipelinerun: ModelPipelineRun) -> bool:
        request = CreateModelPipelineRunRequest()
        request.modelpipelinerun.CopyFrom(modelpipelinerun.raw_message)
        try:
            response = self.__stub.CreateModelPipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, modelpipelinerun: ModelPipelineRun) -> bool:
        request = UpdateModelPipelineRunRequest()
        request.modelpipelinerun.CopyFrom(modelpipelinerun.raw_message)
        try:
            self.__stub.UpdateModelPipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[ModelPipelineRun, bool]:
        request = GetModelPipelineRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetModelPipelineRun(request)
            return ModelPipelineRun(response.modelpipelinerun, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteModelPipelineRunRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteModelPipelineRun(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[ModelPipelineRun], bool]:
        request = ListModelPipelineRunsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListModelPipelineRuns(request)
            return [ModelPipelineRun(item, self) for item in response.modelpipelineruns.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import hashlib
import subprocess
import time
from contextlib import closing
import socket
from typing import List

import grpc
from github.com.metaprov.modelaapi.services.grpcinferenceservice.v1 import grpcinferenceservice_pb2_grpc
from github.com.metaprov.modelaapi.services.grpcinferenceservice.v1.grpcinferenceservice_pb2 import PredictRequest
from github.com.metaprov.modelaapi.services.grpcinferenceservice.v1.grpcinferenceservice_pb2_grpc import \
    GRPCInferenceServiceStub

from modela.ModelaException import ModelaException
from modela.inference.models import PredictionResult


class InferenceService:
    """
    The InferenceService class represents a connection to a gRPC service that exposes the GRPCInferenceService API.
    This API exposes methods to make prediction requests and view information about the current status of the
    prediction proxy.
    """

    def __init__(self, host="localhost", port=None, tls_cert=None, port_forward=False, service_name="", service_namespace=""):
        """
        Connect to the gRPC service.

        :param host: The DNS name or IP that hosts the service.
        :param port: The port which exposes the service.
        :param tls_cert: The TLS certificate of the connection, if connecting through ingress. The Secret containing the
            public key of the ingress can be found in the namespace of the Serving Site that hosts the service.
        """
        if port_forward:
            with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
                s.bind(('', 0))
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                port = s.getsockname()[1]

            self.pf_process = subprocess.Popen("kubectl port-forward svc/%s %d:8080 -n %s" %
                                               (service_name, port, service_namespace),
                                               shell=True, stderr=subprocess.STDOUT)

            time.sleep(1/4)
            tls_cert, host = None, "localhost"

        if tls_cert:
            with open(tls_cert, 'rb') as f:
                credentials = grpc.ssl_channel_credentials(f.read())

            self._channel = grpc.secure_channel(f'{host}', credentials)
        else:
            if port != None:
                self._channel = grpc.insecure_channel(f'{host}:{port}')
            else:
                self._channel = grpc.insecure_channel(f'{host}')

        self._stub = grpcinferenceservice_pb2_grpc.GRPCInferenceServiceStub(self._channel)

    def close(self):
        """Closes Dapr runtime gRPC channel."""
        if self._channel:
            self._channel.close()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if hasattr(self, 'pf_process'):
            self.pf_process.kill()

        self.close()

    def predict(self,
                payload: str,
                validate: bool = True,
                explain: bool = False,
                format: str = "json",
                labeled: bool = False,
                metrics: List[str] = []) -> List[PredictionResult]:
        request = PredictRequest(validate=validate, explain=explain, format=format,
                                 payload=payload, labeled=labeled)
        request.metrics.extend(metrics)
        try:
            response = self._stub.Predict(request)
            return [PredictionResult().copy_from(item) for item in response.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)



import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import DataProductVersion as MDDataProductVersion
from github.com.metaprov.modelaapi.services.dataproductversion.v1.dataproductversion_pb2_grpc import DataProductVersionServiceStub
from github.com.metaprov.modelaapi.services.dataproductversion.v1.dataproductversion_pb2 import CreateDataProductVersionRequest, \
    UpdateDataProductVersionRequest, \
    DeleteDataProductVersionRequest, GetDataProductVersionRequest, ListDataProductVersionsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union

from modela.data.models import DataProductVersionSpec


class DataProductVersion(Resource):
    def __init__(self, item: MDDataProductVersion = MDDataProductVersion(), client=None, namespace="", name="",
                baseline: bool = False, previous_version: str = None):
        """
        :param client: The Data Product Version client repository, which can be obtained through an instance of Modela.
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param baseline: If this version is baseline, then the objects of previous version will be garbage collected.
        :param previous_version: The name of the previous version.
        """
        super().__init__(item, client, namespace=namespace, name=name)

    @property
    def spec(self) -> DataProductVersionSpec:
        return DataProductVersionSpec().copy_from(self._object.spec).set_parent(self._object.spec)

    def default(self):
        DataProductVersionSpec().apply_config(self._object.spec)



class DataProductVersionClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: DataProductVersionServiceStub = stub

    def create(self, dataproductversion: DataProductVersion) -> bool:
        request = CreateDataProductVersionRequest()
        request.dataproductversion.CopyFrom(dataproductversion.raw_message)
        try:
            response = self.__stub.CreateDataProductVersion(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, dataproductversion: DataProductVersion) -> bool:
        request = UpdateDataProductVersionRequest()
        request.dataproductversion.CopyFrom(dataproductversion.raw_message)
        try:
            self.__stub.UpdateDataProductVersion(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[DataProductVersion, bool]:
        request = GetDataProductVersionRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetDataProductVersion(request)
            return DataProductVersion(response.dataproductversion, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteDataProductVersionRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteDataProductVersion(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[DataProductVersion], bool]:
        request = ListDataProductVersionsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListDataProductVersions(request)
            return [DataProductVersion(item, self) for item in response.dataproductversions.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import os

import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import DataSource as MDDataSource
from github.com.metaprov.modelaapi.services.datasource.v1.datasource_pb2_grpc import DataSourceServiceStub
from github.com.metaprov.modelaapi.services.datasource.v1.datasource_pb2 import CreateDataSourceRequest, \
    UpdateDataSourceRequest, \
    DeleteDataSourceRequest, GetDataSourceRequest, ListDataSourceRequest, InferSchemaRequest

from modela.data.common import *
from modela.data.models import ExcelNotebookFormat, CsvFileFormat, Column, DataType, DataSourceSpec, ColumnProfile, \
    DataLocation
from modela.Resource import *
from modela.ModelaException import ModelaException
from typing import List, Union
import pandas
from modela.training.common import *


class DataSource(Resource):
    def __init__(self, item: MDDataSource = MDDataSource(), client=None, namespace="", name="",
                 version=Resource.DefaultVersion,
                 bucket: str = "default-minio-bucket",
                 infer_file: str = None,
                 infer_dataframe: pandas.DataFrame = None,
                 infer_bytes: bytes = None,
                 target_column: str = "",
                 file_type: FlatFileType = None,
                 task_type: TaskType = None,
                 csv_config: CsvFileFormat = None,
                 excel_config: ExcelNotebookFormat = None):
        """
        :param client: The Data Source client repository, which can be obtained through an instance of Modela.
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param version: The version of the resource.
        :param bucket: If data is provided for inference then a bucket must be provided.
        :param infer_file: If specified, the SDK will attempt read a file with the given path and will upload it to
            analyse the columns and generate a schema that will be applied to the resource.
        :param infer_dataframe: If specified, the Pandas DataFrame will be serialized and uploaded to analyse
            the columns and generate a schema that will be applied to the resource.
        :param infer_bytes: If specified, the raw byte data will be uploaded to analyse
            the columns and generate a schema that will be applied to the resource.
        :param target_column: The name of the target column used when training a model. This parameter only has effect
            when data is uploaded to infer a schema.
        :param file_type: The file type of raw data, used when ingesting a Dataset from a file, or creating a data snapshot
            from a database source. If inferring from a dataframe, the file type will default to CSV.
        :param task_type: The target task type in relation to the data being used.
        :param csv_config: The CSV file format of the raw data.
        :param excel_config: The Excel file format of the raw data.
        """
        self.default_resource = False
        super().__init__(item, client, namespace=namespace, name=name, version=version)

        if infer_file is not None:
            with open(infer_file, 'rb') as f:
                infer_bytes = f.read()
            infer_file = os.path.basename(infer_file)
        elif infer_dataframe is not None:
            file_type = FlatFileType.Csv
            infer_bytes = infer_dataframe.to_csv(index=False)
        if infer_bytes is not None:
            infer_location = client.modela.FileService.upload_file(infer_file or name, infer_bytes, client.modela.tenant,
                                                                   namespace, version, bucket, "datasources", name)

            profile = client.infer(namespace, infer_location, data_source=self, version=version)
            schema_columns = self.spec.Schema.Columns
            del schema_columns[:]
            for column in profile:
                col = Column()
                col.Datatype = DataType(column.Type)
                col.Name = column.Name
                col.Enum = column.Values
                if target_column == col.Name:
                    col.Target = True
                    target_column = None

                schema_columns.append(col)

            if target_column != None:
                print("WARNING: The target column {0} was not found in the dataset during schema inference.".format(target_column))

        if self.default_resource:
            self.spec.FileType = file_type
            self.spec.Task = task_type
        if file_type is not None:
            self.spec.FileType = file_type
            if not self.default_resource:
                print("WARNING: Changing the file type of an existing Data Source is not recommended. "
                      "Create a new Data Source resource if necessary.")
        elif self.default_resource:
            print("WARNING: Data Source was created without a file type. Defaulting to FlatFileType.Csv")
            self.spec.FileType = FlatFileType.Csv

        if task_type is not None:
            self.spec.Task = file_type
            if not self.default_resource:
                print("WARNING: Changing the task type of an existing Data Source is not recommended. "
                      "Create a new Data Source resource if necessary.")
        elif self.default_resource:
            print("WARNING: Data Source was created without a task type. Defaulting to TaskType.BinaryClassification")
            self.spec.Task = TaskType.BinaryClassification

        if csv_config is not None:
            self.spec.Csvfile = csv_config

        if excel_config is not None:
            self.spec.ExcelNotebook = excel_config

    @property
    def spec(self) -> DataSourceSpec:
        return DataSourceSpec().copy_from(self._object.spec).set_parent(self._object.spec)

    @property
    def schema(self):
        return self.spec.Schema

    @property
    def target_column(self) -> Column:
        """ Get the target column specified by the schema """
        return [col for col in self.schema.Columns if col.Target][0]

    def column(self, name: str) -> Column:
        """ Get the column with the specified name from the schema """
        search = [col for col in self.schema.Columns if col.Name == name]
        if len(search) == 0:
            raise ValueError("Data source {0} does not have a column named {1}".format(self.name, name))

        return search[0]

    def __getitem__(self, col) -> Column:
        """ Get the column with the specified name from the schema """
        return self.column(col)

    def default(self):
        self.default_resource = True
        DataSourceSpec().apply_config(self._object.spec)


class DataSourceClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: DataSourceServiceStub = stub

    def create(self, datasource: DataSource) -> bool:
        request = CreateDataSourceRequest()
        request.datasource.CopyFrom(datasource.raw_message)
        try:
            response = self.__stub.CreateDataSource(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, datasource: DataSource) -> bool:
        request = UpdateDataSourceRequest()
        request.datasource.CopyFrom(datasource.raw_message)
        try:
            self.__stub.UpdateDataSource(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[DataSource, bool]:
        request = GetDataSourceRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetDataSource(request)
            return DataSource(response.datasource, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteDataSourceRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteDataSource(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[DataSource], bool]:
        request = ListDataSourceRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListDataSources(request)
            return [DataSource(item, self) for item in response.datasources.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def infer(self, namespace: str, location: DataLocation, file_type: FlatFileType = FlatFileType.Csv,
              data_source: DataSource = None, version=Resource.DefaultVersion) -> List[ColumnProfile]:
        request = InferSchemaRequest()
        request.namespace = namespace
        (data_source.spec if data_source else DataSourceSpec()).apply_config(request.datasource.spec)
        request.datasource.spec.versionName = version
        request.datasource.spec.fileType = file_type.value
        request.location.CopyFrom(location.to_message())

        try:
            response = self.__stub.InferSchema(request)
            return [ColumnProfile().copy_from(item) for item in response.columns]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False
